!> @file user_init.f90
!------------------------------------------------------------------------------!
! This file is part of the PALM model system.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2018 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! -----------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_init.f90 2718 2018-01-02 08:49:38Z maronga $
! Corrected "Former revisions" section
! 
! 2696 2017-12-14 17:12:51Z kanani
! Change in file header (GPL part)
!
! 2101 2017-01-05 16:42:31Z suehring
!
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1799 2016-04-05 08:35:55Z gronemeier
! Bugfix: added dots_num to variable list
!
! 1783 2016-03-06 18:36:17Z raasch
! netcdf module name changed + related changes
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute 
!
! 1320 2014-03-20 08:40:49Z raasch
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 211 2008-11-11 04:46:24Z raasch
! Former file user_interface.f90 split into one file per subroutine
!
! Description:
! ------------
!> Execution of user-defined initializing actions
!------------------------------------------------------------------------------!
 SUBROUTINE user_init
 

    USE control_parameters
    
    USE indices
    
    USE kinds
    
    USE netcdf_interface,                                                      &
        ONLY: dots_label, dots_unit, dots_num
    
    USE pegrid
    
    USE user

    IMPLICIT NONE

    CHARACTER (LEN=20) :: field_char   !< 
    INTEGER(iwp) :: dlen !< no. rows of local table
    INTEGER(iwp) ::  jj   !< running index

    IF ( theta_0  == 999. ) THEN
        IF ( TRIM ( theta_file ) == "" ) THEN
            PRINT *, '+++ table_file = "', TRIM ( theta_file ), &
                    '" not found.'
            STOP "end of file"
        ELSE
            OPEN ( 501, FILE=TRIM( theta_file ), FORM='FORMATTED', ACTION='READ')
            dlen = 0
        35  READ ( 501, *, END=36)
            dlen = dlen + 1
            GOTO 35
        36  CONTINUE
            CLOSE( 501 )

            ALLOCATE( WD_time(1:dlen), WD_theta(1:dlen), WD_WD(1:dlen) )
            OPEN ( 501, FILE=TRIM( theta_file ), &
                  FORM='FORMATTED', ACTION='READ', POSITION='REWIND' )
            DO jj = 1,dlen
                READ ( 501, * ) WD_time(jj), WD_WD(jj), WD_theta(jj)
            ENDDO 
            CLOSE( 501 )
        ENDIF
    ENDIF

!-- Oeffnen der Ausgabedatei fuer das Feld der u-Komponente in einer yz-Ebene, das spaeter zur Vorgabe 
!-- eines turbulenten Einstroemens im Falle von nicht-zyklischen Randbedingungen entlang der 
!-- x-Richtung verwendet werden soll                             

    IF ( myid == 0 ) THEN
       OPEN(819, FILE="turbulent_inflow_data_u.dat",FORM="UNFORMATTED",STATUS="REPLACE")          
    END IF   

!-- Oeffnen der Ausgabedatei fuer das Feld der v-Komponente in einer yz-Ebene, das spaeter zur Vorgabe
!-- eines turbulenten Einstroemens im Falle von nicht-zyklischen Randbedingungen entlang der 
!-- x-Richtung verwendet werden soll

    IF ( myid == 0 ) THEN
       OPEN(820, FILE="turbulent_inflow_data_v.dat",FORM="UNFORMATTED",STATUS="REPLACE")
    END IF
 
!-- Oeffnen der Ausgabedatei fuer das Feld der w-Komponente in einer yz-Ebene, das spater zur Vorgabe
!-- eines turbulenten Einstroemens im Falle von nicht-zyklischen Randbedingungen entlang der 
!-- x-Richtung verwendet werden soll

    IF ( myid == 0 ) THEN
       OPEN(821, FILE="turbulent_inflow_data_w.dat",FORM="UNFORMATTED",STATUS="REPLACE")
    END IF

!-- Oeffnen der Ausgabedatei fuer das Feld der subskaligen turbulenten kinetischen Energie, das spaeter
!-- zur Vorgabe eines turbulenten Einstroemens im Falle von nicht-zyklischen Randbedingungen entlang 
!-- der x-Richtung verwendet werden soll

    IF ( myid == 0 ) THEN
       OPEN(822, FILE="turbulent_inflow_data_e.dat",FORM="UNFORMATTED",STATUS="REPLACE")
    END IF

!-- Oeffnen der Ausgabedatei fuer das Feld der potentiellen Temperatur, welches spaeter zur Vorgabe
!-- eines turbulenten Einstroemens im Falle von nicht-zyklischen Randbedingungen entlang der 
!-- x-Richtung verwendet werden soll

    IF ( myid == 0 ) THEN
       OPEN(823, FILE="turbulent_inflow_data_pt.dat",FORM="UNFORMATTED",STATUS="REPLACE")
    END IF

 END SUBROUTINE user_init

